import { createServer } from 'vite'
import fastify from 'fastify'
import { generate } from 'selfsigned'
import request from 'http2-client/lib/request'
import { jest } from '@jest/globals'

const servers = []
let requestManager
updateRequestManager()

export function generateSelfsigned () {
  const pems = generate(null, {
    algorithm: 'sha256',
    days: 30,
    keySize: 2048,
    extensions: [
      // {
      //   name: 'basicConstraints',
      //   cA: true,
      // },
      {
        name: 'keyUsage',
        keyCertSign: true,
        digitalSignature: true,
        nonRepudiation: true,
        keyEncipherment: true,
        dataEncipherment: true
      },
      {
        name: 'extKeyUsage',
        serverAuth: true,
        clientAuth: true,
        codeSigning: true,
        timeStamping: true
      },
      {
        name: 'subjectAltName',
        altNames: [
          {
            // type 2 is DNS
            type: 2,
            value: 'localhost'
          },
          {
            type: 2,
            value: 'localhost.localdomain'
          },
          {
            type: 2,
            value: 'lvh.me'
          },
          {
            type: 2,
            value: '*.lvh.me'
          },
          {
            type: 2,
            value: '[::1]'
          },
          {
            // type 7 is IP
            type: 7,
            ip: '127.0.0.1'
          },
          {
            type: 7,
            ip: 'fe80::1'
          }
        ]
      }
    ]
  })

  // return in a common node http fashion
  return {
    key: pems.private,
    cert: pems.cert
  }
}

export async function closeAllServers () {
  await Promise.all(servers.map(async (server) => {
    try {
      await server.close()
    } catch (err) {
      // do nothing
      console.error(err)
    }
  }))
  servers.splice(0, servers.length)
}

/**
 * @param {import('vite').UserConfig} options
 */
export async function createViteServer (options, port) {
  const server = await createServer(options)
  servers.push(server)
  return server.listen(port)
}

export function createTestServer (config = {}) {
  if (config.https) config.https.allowHTTP1 = true
  config.forceCloseConnections = true
  const server = fastify(config)
  servers.push(server)
  return server
}

export function getPort () {
  return (5000 + Math.random() * 5000) >> 0
}

export function updateRequestManager () {
  requestManager = new request.HttpRequestManager()
}

export function http2Request (url, options = {}) {
  return new Promise((resolve, reject) => {
    options.requestManager = requestManager
    const req = requestManager.request(url, options, (res) => {
      if (res.statusCode !== 200) return reject(res.message || new Error(`Received ${res.statusCode} from http2Request`))

      let buffer = ''
      res
        .on('error', reject)
        .on('data', buf => (buffer += buf.toString()))
        .on('end', () => {
          if (res.headers['content-type']?.startsWith('application/json')) {
            res.data = JSON.parse(buffer)
          } else {
            res.data = buffer
          }
          resolve(res)
        })
    })
    req.on('error', console.error)
    req.end()
  })
}

export function filterCalls (mock, filter) {
  const spy = jest.fn()
  for (const call of mock.mock.calls) {
    if (!filter(...call)) continue
    spy(...call)
  }
  return spy
}

export function filterPingCalls (mock) {
  return filterCalls(mock, (req) => !req.url.startsWith('/ping'))
}
